/* 
  Magnum 
  Copyright (C) 2002-2004 Kaya Memisoglu. All Rights Reserved.

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, under the terms in LICENSE.TXT.

  Kaya Memisoglu  (k.memisoglu@dimajix.de)
*/

#include "models.h"

namespace magnum { namespace sample {
using namespace magnum::graphics;
using namespace magnum::graphics3d;
using namespace magnum::io;

/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
static Texture* createTexture(TextureManager& mgr, const char* name) {
	Texture* tex = NULL;
	
	// Try to load skin
	try {
		tex = mgr.get(String(name));;
		if (tex != NULL) {
			tex->refer();
			return tex;
		}
	} catch(...) {
		tex = NULL;
	}
		
	// Try to cut off path information
	if (strchr(name,'/') == NULL)
		return NULL;
		
	const char* s1 = name;
	const char* s2 = strchr(s1,'/');
	do {
		s1 = s2+1;
		s2 = strchr(s1,'/');
	} while(s2 != NULL);

	try {
		tex = mgr.get(String(s1));;
		if (tex != NULL) {
			tex->refer();
			return tex;
		}
	} catch(...) {
		return NULL;
	}

	return NULL;
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void applySkins(const MeshList& mlist,FileFactory* ff,const char* name) {
	AutoPointer<File> file = ff->open(name, File::ReadOnly);
	if (file == NULL)
		return;
	
	AutoPointer<StreamTokeniser> tokeniser = new StreamTokeniser(file->getInputStream());
	tokeniser->enableSlashSlashComments(true);
	tokeniser->enableSlashStarComments(true);
	tokeniser->setWordChars("01234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_+-.#$!$/\\");
	tokeniser->setOrdinaryChars(",");
	StreamTokeniser::Token token;
	
	// Start parsing skin file
	String meshname, mapname;
	Map<String, Texture*> texmap;
	TextureManager mgr;
	mgr.setFileFactory(ff);
	bool comma = false;
	do {
		token = tokeniser->nextToken();
		if ((token == StreamTokeniser::EndOfLine) || (token == StreamTokeniser::EndOfFile)) {
			if (comma && (meshname.getLength() > 0) && (mapname.getLength() > 0)) {
				Texture* tex = createTexture(mgr, mapname);
				if (tex != NULL)
					texmap.insert(meshname, tex);
			}
			meshname.clear();
			mapname.clear();
			comma = false;
		} else if (token == StreamTokeniser::Ordinary) {
			if (tokeniser->getString()[0] == ',')
				comma = true;
		} else if (token == StreamTokeniser::Quote) {
			if (comma)
				mapname += tokeniser->getString();
			else
				meshname += tokeniser->getString();
		} else if (token == StreamTokeniser::Word) {
			if (comma)
				mapname += tokeniser->getString();
			else
				meshname += tokeniser->getString();
		}
	} while (token != StreamTokeniser::EndOfFile);

	// Set textures in all meshes
	MeshList::Iterator iter;
	AutoPointer<Shader> shader = new TextureShader();
	for (iter = mlist.begin(); iter != mlist.end(); ++iter) {
		Mesh* mesh = *iter;
		Map<String,Texture*>::Iterator tex = texmap.at(String(mesh->getName()));
		if (tex == texmap.end())
			continue;
		Appearance* app = (*iter)->getAppearance();
		if (app == NULL)
			continue;
		app->attachTexture(*tex, 0);
		app->attachShader(shader);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void setShader(const MeshList& meshes, Shader* shader) {
	MeshList::Iterator iter;
	for (iter = meshes.begin(); iter != meshes.end(); iter++) {
		Appearance* app = (*iter)->getAppearance();
		if (app == NULL)
			continue;
		app->attachShader(shader);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void setTexture(const MeshList& meshes, Texture* tex, unsigned int stage) {
	MeshList::Iterator iter;
	for (iter = meshes.begin(); iter != meshes.end(); iter++) {
		Appearance* app = (*iter)->getAppearance();
		if (app == NULL)
			continue;
		app->attachTexture(tex, stage);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void clearTexture(const MeshList& meshes, unsigned int stage) {
	MeshList::Iterator iter;
	for (iter = meshes.begin(); iter != meshes.end(); iter++) {
		Appearance* app = (*iter)->getAppearance();
		if (app == NULL)
			continue;
		app->detachTexture(stage);
	}
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void setTextureMaterial(const MeshList& meshes) {
	Shader* shader = new TextureShader();
	setShader(meshes, shader);
	shader->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void setBumpMaterial(const MeshList& meshes) {
	Shader* shader = new BumpShader();
	setShader(meshes, shader);
	shader->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void setEnvMaterial(const MeshList& meshes) {
	TranslatorJPG j;
	Shader* shader = new EnvironmentShader();
	File* file = SystemFile::open("data\\env.jpg");
	Image* img = j.decode(file->getInputStream());
	Texture* env = Texture2D::create(img, Texture::GenerateAll);
	env->setFilter(Texture::LinearMipmapLinear, Texture::Linear);
	file->release();
	img->release();

	setShader(meshes, shader);
	setTexture(meshes, env, 0);

	shader->release();
	env->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void setEnvBumpMaterial(const MeshList& meshes) {
	TranslatorJPG j;
	Shader* shader = new EnvironmentBumpShader();
	File* file = SystemFile::open("data\\env.jpg");
	Image* img = j.decode(file->getInputStream());
	Texture* env = Texture2D::create(img, Texture::GenerateAll);
	env->setFilter(Texture::LinearMipmapLinear, Texture::Linear);
	file->release();
	img->release();

	setShader(meshes, shader);
	setTexture(meshes, env, 1);

	shader->release();
	env->release();
}


/*----------------------------------------------------------------------------
----------------------------------------------------------------------------*/
void setTexEnvBumpMaterial(const MeshList& meshes) {
	TranslatorJPG j;
	Shader* shader = new TextureEnvironmentBumpShader();
	Texture* tex;
	File* file = SystemFile::open("data\\env.jpg");
	Image* img = j.decode(file->getInputStream());
	Texture* env = Texture2D::create(img, Texture::GenerateAll);
	env->setFilter(Texture::LinearMipmapLinear, Texture::Linear);
	file->release();
	img->release();

	MeshList::Iterator iter;
	for (iter = meshes.begin(); iter != meshes.end(); iter++) {
		Appearance* app = (*iter)->getAppearance();
		if (app == NULL)
			continue;
		app->attachShader(shader);
		tex = app->getTexture(0);
		if (tex != NULL) {
			app->attachTexture(tex,1);
			app->attachTexture(env,2);
		}
	}

	shader->release();
	env->release();
}

}; };
